from .core import ccf, acf
from .dutch import q_ghg, q_glg, q_gvg, ghg, glg, gvg
from .metrics import evp, nse, rmse, sse, avg_dev, rsq, aic, bic
from .tests import runs_test, ljung_box, durbin_watson, stoffer_toloi, \
    diagnostics, plot_acf, plot_diagnostics



"""The following methods may be used to describe the fit between the model
simulation and the observations.

.. autosummary::
   :nosignatures:
   :toctree: ./generated

   rmse
   sse
   avg_dev
   nse
   evp
   rsq
   bic
   aic

Examples
========
These methods may be used as follows:

>>> ps.stats.rmse(sim, obs)

or

>>> ml.stats.rmse()

"""

from numpy import sqrt, log

__all__ = ["rmse", "sse", "avg_dev", "nse", "evp", "rsq", "bic", "aic"]


def rmse(sim=None, obs=None, res=None, missing="drop"):
    """Root mean squared error.

    Parameters
    ----------
    sim: pandas.Series
        Series with the simulated values.
    obs: pandas.Series
        Series with the observed values.
    res: pandas.Series
        Series with the residual values.
    missing: str, optional
        string with the rule to deal with missing values. Only "drop" is
        supported now.

    Notes
    -----
    .. math:: rmse = \\sqrt{\\frac{\\sum{r^2}}{n}}

    where :math:`n` is the number of residuals :math:`r`.

    """
    if res is None:
        res = (sim - obs)

    if missing == "drop":
        res = res.dropna()

    n = res.size
    return sqrt((res ** 2).sum() / n)


def sse(sim, obs, missing="drop"):
    """Sum of the squares of the error (SSE).

    Parameters
    ----------
    sim: pandas.Series
        Series with the simulated values.
    obs: pandas.Series
        Series with the observed values.
    missing: str, optional
        string with the rule to deal with missing values. Only "drop" is
        supported now.

    Notes
    -----
    The SSE is calculated as follows:

    .. math:: sse = \\sum(r^2)

    Where :math:`r` are the residuals.

    """
    res = (sim - obs)

    if missing == "drop":
        res = res.dropna()

    return (res ** 2).sum()


def avg_dev(sim, obs, missing="drop"):
    """Average deviation of the residuals.

    Parameters
    ----------
    sim: pandas.Series
        Series with the simulated values.
    obs: pandas.Series
        Series with the observed values.
    missing: str, optional
        string with the rule to deal with missing values. Only "drop" is
        supported now.

    Notes
    -----
    .. math:: avg_{dev} = \\frac{\\sum(r)}{n}

    where :math:`n` is the number of residuals :math:`r`.

    """
    res = (sim - obs)

    if missing == "drop":
        res = res.dropna()

    return res.mean()


def nse(sim, obs, missing="drop"):
    """Nash-Sutcliffe coefficient for model fit as defined in [nash_1970]_.

    Parameters
    ----------
    sim: pandas.Series
        Series with the simulated values.
    obs: pandas.Series
        Series with the observed values.
    missing: str, optional
        string with the rule to deal with missing values. Only "drop" is
        supported now.

    Notes
    -----
    .. math:: nse = 1 - \\frac{\\sum(h_s-h_o)^2}{\\sum(h_o-\\mu_{h,o})}

    References
    ----------
    .. [nash_1970] Nash, J. E., & Sutcliffe, J. V. (1970). River flow
       forecasting through conceptual models part I-A discussion of
       principles. Journal of hydrology, 10(3), 282-290.

    """
    res = (sim - obs)

    if missing == "drop":
        res = res.dropna()

    ns = 1 - (res ** 2).sum() / ((obs - obs.mean()) ** 2).sum()
    return ns


def evp(sim, obs, missing="drop"):
    """Explained variance percentage according to [asmuth_2012]_.

    Parameters
    ----------
    sim: pandas.Series
        Series with the simulated values.
    obs: pandas.Series
        Series with the observed values.
    missing: str, optional
        string with the rule to deal with missing values. Only "drop" is
        supported now.

    Notes
    -----
    Commonly used goodness-of-fit metric groundwater level models as
    computed in [asmuth_2012]_.

    .. math:: evp = \\frac{\\sigma_h^2 - \\sigma_r^2}{\\sigma_h^2} * 100

    where :math:`\\sigma_h^2` is the variance of the observations and
    :math:`\\sigma_r^2` is the variance of the residuals. The returned value
    is bounded between 0% and 100%.

    References
    ----------
    .. [asmuth_2012] von Asmuth, J., K. Maas, M. Knotters, M. Bierkens,
       M. Bakker, T.N. Olsthoorn, D.G. Cirkel, I. Leunk, F. Schaars, and D.C.
       von Asmuth. 2012. Software for hydrogeologic time series analysis,
       interfacing data with physical insight. Environmental Modelling &
       Software 38: 178–190.

    """
    res = (sim - obs)

    if missing == "drop":
        res = res.dropna()

    if obs.var() == 0.0:
        return 100.
    else:
        ev = max(0.0, 100 * (1 - (res.var(ddof=0) / obs.var(ddof=0))))
    return ev


def rsq(sim, obs, nparam=None, missing="drop"):
    """R-squared, possibly adjusted for the number of free parameters.

    Parameters
    ----------
    sim: pandas.Series
        Series with the simulated values.
    obs: pandas.Series
        Series with the observed values.
    nparam: int, optional
        number of calibrated parameters.
    missing: str, optional
        string with the rule to deal with missing values. Only "drop" is
        supported now.

    Notes
    -----
    .. math:: \\rho_{adj} = 1-  \\frac{n-1}{n-n_{param}}*\\frac{rss}{tss}

    Where n is the number of observations, :math:`n_{param}` the number of
    free parameters, rss the sum of the squared residuals, and tss the total
    sum of squared residuals.

    When nparam is provided, the :math:`\\rho` is
    adjusted for the number of calibration parameters.

    """
    res = (sim - obs)

    if missing == "drop":
        res = res.dropna()

    rss = (res ** 2.0).sum()
    tss = ((obs - obs.mean()) ** 2.0).sum()
    if nparam:
        return 1.0 - (obs.size - 1.0) / (obs.size - nparam) * rss / tss
    else:
        return 1.0 - rss / tss


def bic(sim, obs, nparam, missing="drop"):
    """Bayesian Information Criterium (BIC) according to [akaike_1979]_.

    Parameters
    ----------
    sim: pandas.Series
        Series with the simulated values.
    obs: pandas.Series
        Series with the observed values.
    nparam: int, optional
        number of calibrated parameters.
    missing: str, optional
        string with the rule to deal with missing values. Only "drop" is
        supported now.

    Notes
    -----
    The Bayesian Information Criterium (BIC) is calculated as follows:

    .. math:: BIC = -2 log(L) + n_{param} * log(N)

    where :math:`n_{param}` is the number of calibration parameters.

    References
    ----------
    .. [akaike_1979] Akaike, H. (1979). A Bayesian extension of the minimum
       AIC procedure of autoregressive model fitting. Biometrika, 66(2),
       237-242.

    """
    res = (sim - obs)

    if missing == "drop":
        res = res.dropna()

    return -2.0 * log((res ** 2.0).sum()) + nparam * log(res.size)


def aic(sim, obs, nparam, missing="drop"):
    """Akaike Information Criterium (AIC) according to [akaike_1974]_.

    Parameters
    ----------
    sim: pandas.Series
        Series with the simulated values.
    obs: pandas.Series
        Series with the observed values.
    nparam: int, optional
        number of calibrated parameters.
    missing: str, optional
        string with the rule to deal with missing values. Only "drop" is
        supported now.

    Notes
    -----
    .. math:: AIC = -2 log(L) + 2 nparam

    where :math:`n_{param}` is the number of calibration parameters and L is
    the likelihood function for the model.

    References
    ----------
    .. [akaike_1974] Akaike, H. (1974). A new look at the statistical model
       identification. IEEE transactions on automatic control, 19(6), 716-723.

    """
    res = (sim - obs)

    if missing == "drop":
        res = res.dropna()

    return -2.0 * log((res ** 2.0).sum()) + 2.0 * nparam



"""This files contains the Project class that deals with multiple models at
once.

Warning
-------
This class will soon be deprecated and replaced by a separate Python package
Pastastore that deals with large number of Pastas models. It is strongly
recommended to switch to `Pastastore <https://github.com/pastas/pastastore>`_

Notes
-----
This module is created at Artesia Water by Raoul Collenteur.

.. autosummary::
    :nosignatures:
    :toctree: ./generated

    Project

Example
-------

>>> mls = Project()

"""

from logging import getLogger
from os import getlogin

import numpy as np
import pandas as pd

from .maps import Map
from .plots import Plot
from ..io.base import dump
from ..model import Model
from ..rfunc import Gamma
from ..stressmodels import StressModel2
from ..timeseries import TimeSeries
from ..version import __version__

logger = getLogger(__name__)


class Project:
    """The Project class is a placeholder when multiple time series models
    are analyzed in a batch.

    """

    def __init__(self, name, metadata=None):
        """Initialize a Project instance.

        Parameters
        ----------
        name: str
            Name of the project
        metadata: dict
            Dictionary with any metadata information on the project.

        """
        logger.warning("DeprecationWarning: The Project class will be "
                       "deprecated in a future version of Pastas (end of "
                       "2020)! Consider switching to pastastore: "
                       "https://github.com/pastas/pastastore.")
        self.models = {}
        self.name = name

        # DataFrames to store the data of the oseries and stresses
        columns = ["name", "series", "kind", "x", "y", "z"]
        self.stresses = pd.DataFrame(columns=columns)
        self.oseries = pd.DataFrame(columns=columns)

        # Project metadata and file information
        self.metadata = self.get_metadata(metadata)
        self.file_info = self.get_file_info()

        # Load other modules
        self.plots = Plot(self)
        self.maps = Map(self)

    def add_series(self, series, name=None, kind=None, metadata=None,
                   settings=None, **kwargs):
        """Internal method to add series to the oseries or stresses database.

        Parameters
        ----------
        series: pandas.Series / pastas.TimeSeries
            Series object.
        name: str
            String with the name of the series that will be maintained in
            the database.
        kind: str
            The kind of series that is added. When oseries are added it is
            necessary to state "oseries" here.
        metadata: dict
            Dictionary with any metadata that will be passed to the
            TimeSeries object that is created internally.
        settings: dict or str
            Dictionary with any settings that will be passed to the
            TimeSeries object that is created internally.

        """
        if name is None:
            name = series.name

        if not isinstance(name, str):
            name = str(name)

        series.name = name

        if kind == "oseries":
            data = self.oseries
            if settings is None:
                settings = "oseries"
        else:
            data = self.stresses

        if name in data.index:
            warning = ("Time series with name {} is already present in the "
                       "database. Existing series is overwitten.").format(name)
            logger.warning(warning)
        try:
            ts = TimeSeries(series=series, name=name, settings=settings,
                            metadata=metadata, **kwargs)
        except:
            logger.error("An error occurred. Time series {} is omitted "
                         "from the database.".format(name))
            return

        data.at[name, "name"] = name
        data.at[name, "series"] = ts  # Do not add as first!
        data.at[name, "kind"] = kind

        # Transfer the metadata (x, y and z) to dataframe as well to increase speed.
        for i in ts.metadata.keys():
            value = ts.metadata[i]
            data.loc[name, i] = value

    def add_oseries(self, series, name=None, metadata=None, settings="oseries",
                    **kwargs):
        """Convenience method to add oseries to project

        Parameters
        ----------
        series: pandas.Series / pastas.TimeSeries
            Series object.
        name: str
            String with the name of the series that will be maintained in
            the database.
        metadata: dict
            Dictionary with any metadata that will be passed to the
            TimeSeries object that is created internally.
        settings: dict or str
            Dictionary with any settings that will be passed to the
            TimeSeries object that is created internally.

        Returns
        -------

        """
        self.add_series(series, name=name, metadata=metadata,
                        settings=settings, kind="oseries", **kwargs)

    def add_stress(self, series, name=None, kind=None, metadata=None,
                   settings=None, **kwargs):
        """Convenience method to add stress series to project

        Parameters
        ----------
        series: pandas.Series / pastas.TimeSeries
            Series object.
        name: str
            String with the name of the series that will be maintained in
            the database.
        kind: str
            The kind of series that is added. When oseries are added it is
            necessary to state "oseries" here.
        metadata: dict
            Dictionary with any metadata that will be passed to the
            TimeSeries object that is created internally.
        settings: dict or str
            Dictionary with any settings that will be passed to the
            TimeSeries object that is created internally.

        """
        self.add_series(series, name=name, metadata=metadata,
                        settings=settings, kind=kind, **kwargs)

    def add_model(self, oseries, model_name=None, **kwargs):
        """Method to add a Pastas Model instance based on one of the oseries.

        Parameters
        ----------
        oseries: str
            string with the exact names of one of the oseries indices.
        model_name: str
            Name of the model
        kwargs: dict
            any arguments that are taken by the Pastas Model instance can be
            provided.

        Returns
        -------
        ml: pastas.Model
            Pastas Model generated with the oseries and arguments provided.

        """
        if model_name is None:
            model_name = oseries

        # Validate name and ml_name before continuing
        if model_name in self.models.keys():
            warning = ("Model name {} is not unique, existing model is "
                       "overwritten.").format(model_name)
            logger.warning(warning)
        if oseries not in self.oseries.index:
            error = ("Oseries name {} is not present in the database. Make "
                     "sure to provide a valid name.").format(model_name)
            logger.error(error)
            return

        oseries = self.oseries.loc[oseries, "series"]
        ml = Model(oseries, name=model_name, **kwargs)

        # Add new model to the models dictionary
        self.models[model_name] = ml

        return ml

    def add_models(self, oseries="all", model_name_prefix="",
                   model_name_suffix="", **kwargs):
        """Method to add multiple Pastas Model instances based on one
        or more of the oseries.

        Parameters
        ----------
        oseries: str or list, optional
            names of the oseries, if oseries is "all" all series in self.series
            are used
        model_name_prefix: str, optional
            prefix to use for model names
        model_name_suffix: str, optional
            suffix to use for model names
        kwargs: dict
            any arguments that are taken by the Pastas Model instance can be
            provided.

        Returns
        -------
        mls: list of str
            list of modelnames corresponding to the keys in the self.models
            dictionary

        """

        if oseries == "all":
            oseries_list = self.oseries.index
        elif isinstance(oseries, str):
            oseries_list = [oseries]
        elif isinstance(oseries, list):
            oseries_list = oseries

        mls = []
        for oseries_name in oseries_list:
            model_name = model_name_prefix + oseries_name + model_name_suffix

            # Add new model
            ml = self.add_model(oseries_name, model_name, **kwargs)
            mls.append(ml.name)

        return mls

    def add_recharge(self, mls=None, rfunc=Gamma, name="recharge", **kwargs):
        """Add a StressModel2 to the time series models. The
        selection of the precipitation and evaporation time series is based
        on the shortest distance to the a stresses in the stresses list.

        Parameters
        ----------
        mls: list of str, optional
            list of model names, if None all models in project are
            used.
        rfunc: pastas.rfunc, optional
            response function, default is the Gamma function.
        name: str, optional
            name of the stress, default is "recharge".
        **kwargs:
            arguments are pass to the StressModel2 function

        Notes
        -----
        The behaviour of this method will change in the future to add a
        RechargeModel instead of StressModel2.


        """
        if mls is None:
            mls = self.models.keys()
        elif isinstance(mls, Model):
            mls = [mls.name]

        for mlname in mls:
            ml = self.models[mlname]
            oseries = ml.oseries.name
            prec_name = \
                self.get_nearest_stresses(oseries, kind="prec").iloc[0][0]
            prec = self.stresses.loc[prec_name, "series"]
            evap_name = \
                self.get_nearest_stresses(oseries, kind="evap").iloc[0][0]
            evap = self.stresses.loc[evap_name, "series"]

            recharge = StressModel2([prec, evap], rfunc, name=name, **kwargs)

            ml.add_stressmodel(recharge)

    def del_oseries(self, name):
        """Method that safely removes oseries from the project. It validates
        that the oseries is not used in any model.

        Parameters
        ----------
        name: str
            string with a single oseries name.

        """
        if name not in self.oseries.index:
            error = ("Time series with name {} is not present in the database."
                     " Please provide a different name.").format(name)
            logger.error(error)
        else:
            self.oseries.drop(name, inplace=True)
            logger.info(
                "Oseries with name {} deleted from the database.".format(name))

    def del_stress(self, name):
        """Method that removes oseries from the project.

        Parameters
        ----------
        name: list or str
            list with multiple or string with a single stress name.

        """
        if name not in self.stresses.index:
            error = ("Stress with name {} is not present in the database."
                     " Please provide a different name.").format(name)
            logger.error(error)
        else:
            self.stresses.drop(name, inplace=True)
            logger.info(
                "Stress with name {} deleted from the database.".format(name))

    def del_model(self, ml_name):
        """Method to safe-delete a model from the project.

        Parameters
        ----------
        ml_name: str
            String with the model name.

        """
        name = self.models.pop(ml_name, None)
        info = "Model with name {} deleted from the database.".format(name)
        logger.info(info)

    def update_model_series(self):
        """Update all the Model series by their originals in self.oseries and
        self.stresses. This can for example be useful when new data is
        added to any of the series in mls.oseries and mls.stresses

        """
        for ml in self.models.values():
            oname = ml.oseries.name
            ml.oseries.series_original = self.oseries.loc[
                oname, "series"].series_original
            for sm in ml.stressmodels:
                for st in ml.stressmodels[sm].stress:
                    st.series_original = self.stresses.loc[
                        st.name, "series"].series_original
            # set oseries_calib empty, so it is determined again the next time
            ml.oseries_calib = None

    def solve_models(self, mls=None, report=False, ignore_solve_errors=False,
                     verbose=False, **kwargs):
        """Solves the models in mls

        mls: list of str, optional
            list of model names, if None all models in the project are solved.
        report: boolean, optional
            determines if a report is printed when the model is solved.
        ignore_solve_errors: boolean, optional
            if True errors emerging from the solve method are ignored.
        **kwargs:
            arguments are passed to the solve method.

        """
        if mls is None:
            mls = self.models.keys()
        elif isinstance(mls, Model):
            mls = [mls.name]

        for ml_name in mls:
            if verbose:
                print("solving model -> {}".format(ml_name))

            ml = self.models[ml_name]

            m_kwargs = {}
            for key, value in kwargs.items():
                if isinstance(value, pd.Series):
                    m_kwargs[key] = value.loc[ml_name]
                else:
                    m_kwargs[key] = value
            # Convert timestamps
            for tstamp in ["tmin", "tmax"]:
                if tstamp in m_kwargs:
                    m_kwargs[tstamp] = pd.Timestamp(m_kwargs[tstamp])

            try:
                ml.solve(report=report, **m_kwargs)
            except Exception as e:
                if ignore_solve_errors:
                    warning = "solve error ignored for -> {}".format(ml.name)
                    logger.warning(warning)
                else:
                    raise e

    def get_nearest_stresses(self, oseries=None, stresses=None, kind=None,
                             n=1):
        """Method to obtain the nearest (n) stresses of a specific kind.

        Parameters
        ----------
        oseries: str
            String with the name of the oseries
        stresses: str or list of str
            String with the name of the stresses
        kind:
            String with the name of the stresses
        n: int
            Number of stresses to obtain

        Returns
        -------
        stresses:
            List with the names of the stresses.

        """

        distances = self.get_distances(oseries, stresses, kind)

        data = pd.DataFrame(columns=np.arange(n))

        for series in distances.index:
            series = pd.Series(distances.loc[series].sort_values().index[:n],
                               name=series)
            data = data.append(series)

        return data

    def get_distances(self, oseries=None, stresses=None, kind=None):
        """Method to obtain the distances in meters between the stresses and
        oseries.

        Parameters
        ----------
        oseries: str or list of str
        stresses: str or list of str
        kind: str

        Returns
        -------
        distances: pandas.DataFrame
            Pandas DataFrame with the distances between the oseries (index)
            and the stresses (columns).

        """
        if isinstance(oseries, str):
            oseries = [oseries]
        elif oseries is None:
            oseries = self.oseries.index

        if stresses is None and kind is None:
            stresses = self.stresses.index
        elif stresses is None:
            stresses = self.stresses[self.stresses.kind == kind].index
        elif stresses is not None and kind is not None:
            mask = self.stresses.kind == kind
            stresses = self.stresses.loc[stresses].loc[mask].index

        xo = pd.to_numeric(self.oseries.loc[oseries, "x"])
        xt = pd.to_numeric(self.stresses.loc[stresses, "x"])
        yo = pd.to_numeric(self.oseries.loc[oseries, "y"])
        yt = pd.to_numeric(self.stresses.loc[stresses, "y"])

        xh, xi = np.meshgrid(xt, xo)
        yh, yi = np.meshgrid(yt, yo)

        distances = pd.DataFrame(np.sqrt((xh - xi) ** 2 + (yh - yi) ** 2),
                                 index=oseries, columns=stresses)

        return distances

    def get_parameters(self, parameters, models=None, param_value="optimal"):
        """Method to get the parameters from each model. NaN-values are
        returned when the parameters is not present in the model.

        Parameters
        ----------
        parameters: list
            List with the names of the parameters. The parameter does not
            have to be used in all models.
        models: list
            List with the names of the models. These have to be in the
            Project models dictionary.
        param_value: str
            String with the parameter value that needs to be collected:
            Options are: initial, optimal (default), pmax, pmin and vary.

        Returns
        -------
        data: pandas.DataFrame or pandas.Series
            Returns a pandas DataFrame with the models name as the index and
            the parameters as columns. A pandas Series is returned when only
            one parameter values is collected.

        """
        if models is None:
            models = self.models.keys()

        data = pd.DataFrame(index=models, columns=parameters)

        for ml_name in models:
            ml = self.models[ml_name]
            for parameter in parameters:
                if parameter in ml.parameters.index:
                    value = ml.parameters.loc[parameter, param_value]
                    data.loc[ml_name, parameter] = value

        data = data.squeeze()
        return data.astype(float)

    def get_statistics(self, statistics, models=None, **kwargs):
        """Method to get the statistics for each model.

        Parameters
        ----------
        statistics: list
            List with the names of the statistics to calculate for each model.
        models: list
            List with the names of the models. These have to be in the
            Project models dictionary.

        Returns
        -------
        data: pandas.DataFrame or pandas.Series

        """
        if models is None:
            models = self.models.keys()

        data = pd.DataFrame(index=models, columns=statistics)

        for ml_name in models:
            ml = self.models[ml_name]
            for statistic in statistics:
                value = ml.stats.__getattribute__(statistic)(**kwargs)
                data.loc[ml_name, statistic] = value

        data = data.squeeze()
        return data.astype(float)

    def get_locations_geodataframe(self, models=None, **kwargs):
        import geopandas as gpd
        from shapely.geometry import Point

        if models is None:
            models = self.models.keys()

        data = pd.DataFrame(index=models)

        data = data.join(self.oseries.loc[models, ["x", "y", "z"]])
        data["geometry"] = [Point(xy[0], xy[1]) for xy in
                            self.oseries.loc[models, ["x", "y"]].values]
        data = gpd.GeoDataFrame(data, geometry="geometry", **kwargs)
        return data

    def get_oseries_metadata(self, oseries, metadata):
        """Method to get the metadata for all oseries.

        Parameters
        ----------
        oseries: list
            list with the oseries.
        metadata: list
            list with the metadata keywords to obtain.

        Returns
        -------
        data: pandas.DataFrame

        """
        data = pd.DataFrame(data=None, index=oseries, columns=metadata)

        for oseries in data.index:
            meta = self.oseries.loc[oseries, "series"].metadata
            for key in metadata:
                data.loc[oseries, key] = meta[key]

        return data

    def get_oseries_settings(self, oseries, settings):
        """Method to obtain the settings from each oseries TimeSeries object.

        Parameters
        ----------
        oseries
        settings

        Returns
        -------
        data: pandas.DataFrame
            Pandas DataFrame with the oseries as index, settings as columns
            and the values as data.

        """
        data = pd.DataFrame(data=None, index=oseries, columns=settings)

        for oseries in data.index:
            sets = self.oseries.loc[oseries, "series"].settings
            for key in settings:
                data.loc[oseries, key] = sets[key]

        return data

    @staticmethod
    def get_metadata(meta=None):
        metadata = {"projection": None}
        if meta:
            metadata.update(meta)

        return metadata

    @staticmethod
    def get_file_info():
        file_info = {
            "date_created": pd.Timestamp.now(),
            "date_modified": pd.Timestamp.now(),
            "pastas_version": __version__,
        }
        try:
            file_info["owner"] = getlogin()
        except:
            file_info["owner"] = "Unknown"
        return file_info

    def to_file(self, fname, **kwargs):
        """Method to write a Pastas project to a file.

        Parameters
        ----------
        fname: str

        """
        data = self.to_dict(**kwargs)
        return dump(fname, data)

    def to_dict(self, series=False):
        """Internal method to export a Pastas Project as a dictionary.

        Parameters
        ----------
        series: bool, optional
            export model input-series when True. Only export the name of
            the model input_series when False

        Returns
        -------
        data: dict
            A dictionary with all the project data

        """
        data = {
            "name": self.name,
            "metadata": self.metadata,
            "file_info": self.file_info,
            "oseries": self.series_to_dict(self.oseries),
            "stresses": self.series_to_dict(self.stresses),
            "models": {}
        }

        # Add Models
        for name, ml in self.models.items():
            data["models"][name] = ml.to_dict(series=series, file_info=False)

        return data

    @staticmethod
    def series_to_dict(series):
        """Internal method used to export the time series."""
        series = series.to_dict(orient="index")

        for name in series.keys():
            ts = series[name]["series"]
            series[name]["series"] = ts.to_dict(series=True)

        return series



"""The following methods may be used to describe the fit between the model
simulation and the observations.

.. currentmodule:: pastas.modelstats.Statistics

.. autosummary::
   :nosignatures:
   :toctree: ./generated

   summary
   many
   all

Examples
========
These methods may be used as follows.


    >>> ml.stats.summary()
                                         Value
    Statistic
    Pearson R^2                       0.87
    Root mean squared error           0.43
    Bayesian Information Criterion    113.
    Average Deviation                 0.33
    Explained variance percentage     72.7
    Akaike InformationCriterion       25.3

"""

from numpy import nan
from pandas import DataFrame

from .decorators import model_tmin_tmax
from .stats import metrics, diagnostics


class Statistics:
    # Save all statistics that can be calculated.
    ops = {'evp': 'Explained variance percentage',
           'rmse': 'Root mean squared error',
           'rmsn': 'Root mean squared noise',
           'sse': 'Sum of squares of the error',
           'avg_dev': 'Average Deviation',
           'rsq': 'Pearson R^2',
           'rsq_adj': 'Adjusted Pearson R^2',
           'bic': 'Bayesian Information Criterion',
           'aic': 'Akaike Information Criterion',
           'nse': 'Nash-Sutcliffe Efficiency'}

    def __init__(self, ml):
        """This class provides statistics to to pastas Model class.

        Parameters
        ----------
        ml: Pastas.model.Model
            ml is a time series Model that is calibrated.

        Notes
        -----
        To obtain a list of all statistics that are included type:

        >>> print(ml.stats.ops)

        """
        # Save a reference to the model.
        self.ml = ml

    def __repr__(self):
        msg = """This module contains all the statistical functions that are
included in Pastas. To obtain a list of all statistics that are included type:

    >>> print(ml.stats.ops)"""
        return msg

    @model_tmin_tmax
    def rmse(self, tmin=None, tmax=None):
        """Root mean squared error of the residuals.

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional

        See Also
        --------
        pastas.stats.rmse

        """
        sim = self.ml.simulate(tmin=tmin, tmax=tmax)
        obs = self.ml.observations(tmin=tmin, tmax=tmax)
        return metrics.rmse(sim=sim, obs=obs)

    @model_tmin_tmax
    def rmsn(self, tmin=None, tmax=None):
        """Root mean squared error of the noise.

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional

        Returns
        -------
        float or nan
            Return a float if noisemodel is present, nan if not.

        See Also
        --------
        pastas.stats.rmse

        """
        if not self.ml.settings["noise"]:
            return nan
        else:
            res = self.ml.noise(tmin=tmin, tmax=tmax)
            return metrics.rmse(res=res)

    @model_tmin_tmax
    def sse(self, tmin=None, tmax=None):
        """Sum of the squares of the error (SSE)

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional

        See Also
        --------
        pastas.stats.sse

        """
        sim = self.ml.simulate(tmin=tmin, tmax=tmax)
        obs = self.ml.observations(tmin=tmin, tmax=tmax)
        return metrics.sse(sim=sim, obs=obs)

    @model_tmin_tmax
    def avg_dev(self, tmin=None, tmax=None):
        """Average deviation of the residuals.

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional

        See Also
        --------
        pastas.stats.avg_dev

        """
        sim = self.ml.simulate(tmin=tmin, tmax=tmax)
        obs = self.ml.observations(tmin=tmin, tmax=tmax)
        return metrics.avg_dev(sim=sim, obs=obs)

    @model_tmin_tmax
    def nse(self, tmin=None, tmax=None):
        """Nash-Sutcliffe coefficient for model fit .

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional

        See Also
        --------
        pastas.stats.nse

        """
        sim = self.ml.simulate(tmin=tmin, tmax=tmax)
        obs = self.ml.observations(tmin=tmin, tmax=tmax)
        return metrics.nse(sim=sim, obs=obs)

    @model_tmin_tmax
    def evp(self, tmin=None, tmax=None):
        """Explained variance percentage.

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional

        See Also
        --------
        pastas.stats.evp

        """
        sim = self.ml.simulate(tmin=tmin, tmax=tmax)
        obs = self.ml.observations(tmin=tmin, tmax=tmax)
        return metrics.evp(sim=sim, obs=obs)

    @model_tmin_tmax
    def rsq(self, tmin=None, tmax=None):
        """R-squared.

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional

        See Also
        --------
        pastas.stats.rsq

        """
        sim = self.ml.simulate(tmin=tmin, tmax=tmax)
        obs = self.ml.observations(tmin=tmin, tmax=tmax)
        return metrics.rsq(sim=sim, obs=obs)

    @model_tmin_tmax
    def rsq_adj(self, tmin=None, tmax=None):
        """R-squared Adjusted for the number of free parameters.

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional

        See Also
        --------
        pastas.stats.rsq

        """
        nparam = self.ml.parameters.index.size
        sim = self.ml.simulate(tmin=tmin, tmax=tmax)
        obs = self.ml.observations(tmin=tmin, tmax=tmax)
        return metrics.rsq(sim=sim, obs=obs, nparam=nparam)

    @model_tmin_tmax
    def bic(self, tmin=None, tmax=None):
        """Bayesian Information Criterium (BIC).

        Parameters
        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional

        See Also
        --------
        pastas.stats.bic

        """
        nparam = self.ml.parameters.index.size
        sim = self.ml.simulate(tmin=tmin, tmax=tmax)
        obs = self.ml.observations(tmin=tmin, tmax=tmax)
        return metrics.bic(sim=sim, obs=obs, nparam=nparam)

    @model_tmin_tmax
    def aic(self, tmin=None, tmax=None):
        """Akaike Information Criterium (AIC).

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional

        See Also
        --------
        pastas.stats.rsq

        """
        nparam = self.ml.parameters.index.size
        sim = self.ml.simulate(tmin=tmin, tmax=tmax)
        obs = self.ml.observations(tmin=tmin, tmax=tmax)
        return metrics.aic(sim=sim, obs=obs, nparam=nparam)

    @model_tmin_tmax
    def summary(self, tmin=None, tmax=None, stats='basic'):
        """Prints a summary table of the model statistics.

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional
        stats : str or dict
            dictionary of the desired statistics or a string with one of the
            predefined sets. Supported options are: 'basic', 'all', and 'dutch'

        Returns
        -------
        stats : Pandas.DataFrame
            single-column DataFrame with calculated statistics

        Notes
        -----
        The set of statistics that are printed are stats by a dictionary of
        the desired statistics.

        """
        output = {
            'basic': {
                'evp': 'Explained variance percentage',
                'rmse': 'Root mean squared error',
                'avg_dev': 'Average Deviation',
                'rsq': 'Pearson R^2',
                'bic': 'Bayesian Information Criterion',
                'aic': 'Akaike Information Criterion'},
        }

        # get labels and method names for stats output
        if stats == 'all':
            # sort by key, label, method name
            selected_output = sorted([(k, l, f) for k, d in output.items()
                                      for f, l in d.items()])
        else:
            # sort by name, method name
            selected_output = sorted([(0, l, f) for f, l in
                                      output[stats].items()])

        # compute statistics
        labels_and_values = [(l, getattr(self, f)(tmin=tmin, tmax=tmax))
                             for _, l, f in selected_output]
        labels, values = zip(*labels_and_values)

        stats = DataFrame(index=list(labels), data=list(values),
                          columns=['Value'])
        stats.index.name = 'Statistic'
        return stats

    @model_tmin_tmax
    def many(self, tmin=None, tmax=None, stats=None):
        """This method returns the values for a provided list of statistics.

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional
        stats: list, optional
            list of statistics that need to be calculated.

        Returns
        -------
        data: pandas.DataFrame

        """
        if stats is None:
            stats = ['evp', 'rmse', 'rmsn', 'rsq']

        data = DataFrame(index=[0], columns=stats)
        for k in stats:
            data.iloc[0][k] = (getattr(self, k)(tmin=tmin, tmax=tmax))

        return data

    @model_tmin_tmax
    def all(self, tmin=None, tmax=None):
        """Returns a dictionary with all the statistics.

        Parameters
        ----------
        tmin: str or pandas.Timestamp, optional
        tmax: str or pandas.Timestamp, optional

        Returns
        -------
        stats: pd.DataFrame
            Dataframe with all possible statistics

        """
        stats = DataFrame(columns=['Value'])
        for k in self.ops.keys():
            stats.loc[k] = (getattr(self, k)(tmin=tmin, tmax=tmax))

        return stats

    @model_tmin_tmax
    def diagnostics(self, tmin=None, tmax=None, alpha=0.05, stats=(),
                    float_fmt="{0:.2f}"):
        if self.ml.noisemodel and self.ml.settings["noise"]:
            series = self.ml.noise(tmin=tmin, tmax=tmax)
            nparam = self.ml.noisemodel.nparam
        else:
            series = self.ml.residuals(tmin=tmin, tmax=tmax)
            nparam = 0

        return diagnostics(series=series, alpha=alpha, nparam=nparam,
                           stats=stats, float_fmt=float_fmt)



"""The following noise models are available in Pastas. A Noise model may be
used to transform the residual series into a noise series that better
represents white noise.

Supported Noise Models
----------------------

.. autosummary::
    :nosignatures:
    :toctree: ./generated

    NoiseModel
    ArmaModel

Examples
--------
By default, a noise model is added to a Pastas model. It is possible to
replace the default model with different models as follows:

>>> n = ps.ArmaModel()
>>> ml.add_noisemodel(n)

or, to delete the noise model from the model:

>>> ml.del_noisemodel()

See Also
--------
pastas.model.Model.add_noisemodel

"""

import numpy as np
from pandas import Timedelta, DataFrame, Series

from .decorators import set_parameter, njit

__all__ = ["NoiseModel", "NoiseModelOld", "ArmaModel"]


class NoiseModelBase:
    _name = "NoiseModelBase"

    def __init__(self):
        self.nparam = 1
        self.name = "noise"
        self.parameters = DataFrame(
            columns=["initial", "pmin", "pmax", "vary", "name"])

    def set_init_parameters(self, oseries=None):
        if oseries is not None:
            pinit = oseries.index.to_series().diff() / Timedelta(1, "D")
            pinit = pinit.median()
        else:
            pinit = 14.0
        self.parameters.loc["noise_alpha"] = (pinit, 1e-5, 5000, True, "noise")

    @set_parameter
    def set_initial(self, name, value):
        """
        Internal method to set the initial parameter value.

        Notes
        -----
        The preferred method for parameter setting is through the model.

        """
        self.parameters.loc[name, "initial"] = value

    @set_parameter
    def set_pmin(self, name, value):
        """
        Internal method to set the minimum value of the noisemodel.

        Notes
        -----
        The preferred method for parameter setting is through the model.

        """
        self.parameters.loc[name, "pmin"] = value

    @set_parameter
    def set_pmax(self, name, value):
        """
        Internal method to set the maximum parameter values.

        Notes
        -----
        The preferred method for parameter setting is through the model.

        """
        self.parameters.loc[name, "pmax"] = value

    @set_parameter
    def set_vary(self, name, value):
        """
        Internal method to set if the parameter is varied.

        Notes
        -----
        The preferred method for parameter setting is through the model.

        """
        self.parameters.loc[name, "vary"] = value

    def to_dict(self):
        return {"type": self._name}

    @staticmethod
    def weights(res, parameters):
        return 1


class NoiseModel(NoiseModelBase):
    """
    Noise model with exponential decay of the residuals and weighting.
    Differences compared to NoiseModelOld:

    1. First value is residual
    2. First weight is 1 / sig_residuals (i.e., delt = infty)
    3. Normalization of weights as in Von Asmuth and Bierkens (2005), optional

    Parameters
    ----------
    norm: boolean
        Boolean to indicate whether weights are normalized according to
        the Von Asmuth and Bierkens (2005) paper.

    Notes
    -----
    Calculates the noise [1]_ according to:

    .. math::

        v(t_1) = r(t_1) - r(t_0) * \\exp(- \\Delta t / \\alpha)

    Calculates the weights as

    .. math::

        w = 1 / \\sqrt{(1 - \\exp(-2 \\Delta t / \\alpha))}

    The units of the alpha parameter is always in days.

    References
    ----------
    .. [1] von Asmuth, J. R., and M. F. P. Bierkens (2005), Modeling
           irregularly spaced residual series as a continuous stochastic
           process, Water Resour. Res., 41, W12404, doi:10.1029/2004WR003726.

    """
    _name = "NoiseModel"

    def __init__(self, norm=True):
        NoiseModelBase.__init__(self)
        self.norm = norm
        self.nparam = 1
        self.set_init_parameters()

    @staticmethod
    def simulate(res, parameters):
        """
        Simulate noise from the residuals.

        Parameters
        ----------
        res: pandas.Series
            The residual series.
        parameters: array-like
            Alpha parameters used by the noisemodel.

        Returns
        -------
        noise: pandas.Series
            Series of the noise.

        """
        alpha = parameters[0]
        odelt = (res.index[1:] - res.index[:-1]).values / Timedelta("1d")
        v = Series(index=res.index, dtype="float64", name="Noise")
        v.iloc[0] = res.values[0]
        v.iloc[1:] = res.values[1:] - np.exp(-odelt / alpha) * res.values[:-1]
        return v

    def weights(self, res, parameters):
        """
        Method to calculate the weights for the noise.

        Parameters
        ----------
        res: pandas.Series
            Pandas Series with the residuals to compute the weights for. The
            Series index must be a DatetimeIndex.
        parameters: numpy.ndarray
            numpy array with the parameters used in the noise mdoel.

        Returns
        -------
        w: pandas.Series
            Series of the weights.

        Notes
        -----
        Weights are

        .. math::

        w = 1 / sqrt((1 - exp(-2 \\Delta t / \\alpha)))

        which are then normalized so that sum(w) = len(res)

        """
        alpha = parameters[0]
        odelt = np.empty(res.size)
        odelt[0] = 1e12  # large for first measurement
        odelt[1:] = (res.index[1:] - res.index[:-1]).values / Timedelta("1d")
        exp = np.exp(-2.0 / alpha * odelt)  # Twice as fast as 2*odelt/alpha
        # weights of noise, not noise^2
        w = Series(data=1 / np.sqrt(1.0 - exp), index=res.index,
                   dtype="float64", name="noise_weights")
        if self.norm:
            w = w.multiply(np.exp(1.0 / (2.0 * odelt.size) *
                                  np.sum(np.log(1.0 - exp))))
        return w


class NoiseModelOld(NoiseModelBase):
    """
    Noise model with exponential decay of the residual and weighting.

    Notes
    -----
    Calculates the noise [1]_ according to:

    .. math::

        v(t1) = r(t1) - r(t0) * exp(- (\\frac{\\Delta t}{\\alpha})

    Note that in the referenced paper, alpha is defined as the inverse of
    alpha used in Pastas. The unit of the alpha parameter is always in days.

    References
    ----------
    .. [1] von Asmuth, J. R., and M. F. P. Bierkens (2005), Modeling
           irregularly spaced residual series as a continuous stochastic
           process, Water Resour. Res., 41, W12404, doi:10.1029/2004WR003726.

    """
    _name = "NoiseModel"

    def __init__(self):
        NoiseModelBase.__init__(self)
        self.nparam = 1
        self.set_init_parameters()

    def simulate(self, res, parameters):
        """
        Simulate noise from the residuals.

        Parameters
        ----------
        res: pandas.Series
            The residual series.
        parameters: array-like
            Alpha parameters used by the noisemodel.

        Returns
        -------
        noise: pandas.Series
            Series of the noise.

        """
        alpha = parameters[0]
        odelt = (res.index[1:] - res.index[:-1]).values / Timedelta("1d")
        # res.values is needed else it gets messed up with the dates
        res.iloc[1:] -= np.exp(-odelt / alpha) * res.values[:-1]
        res.name = "Noise"
        return res

    @staticmethod
    def weights(res, parameters):
        """
        Method to calculate the weights for the noise.

        Based on the sum of weighted squared noise (SWSI) method.

        Parameters
        ----------
        res: pandas.Series
            The residual series.
        parameters: array-like
            Alpha parameters used by the noisemodel.

        Returns
        -------
        w: numpy.ndarray
            Array with the weights.

        """
        alpha = parameters[0]
        odelt = (res.index[1:] - res.index[:-1]).values / Timedelta("1d")
        # divide power by 2 as nu / sigma is returned
        power = 1.0 / (2.0 * odelt.size)
        exp = np.exp(-2.0 / alpha * odelt)  # Twice as fast as 2*odelt/alpha
        w = np.exp(power * np.sum(np.log(1.0 - exp))) / np.sqrt(1.0 - exp)
        w = np.insert(w, 0, 0)  # Set first weight to zero
        w = Series(w, res.index)
        return w


class ArmaModel(NoiseModelBase):
    """
    ARMA(1,1) Noise model to simulate the noise.

    Notes
    -----
    Calculates the noise according to:

    .. math::

        \\upsilon_t = r_t - r_{t-1} e^{-\\Delta t/\\alpha} - \\upsilon_{t-1}
        e^{-\\Delta t/\\beta}

    The unit of the alpha parameter is always in days.

    Warnings
    --------
    This model has only been tested on regular time steps and should not be
    used for irregular time steps yet.

    """
    _name = "ArmaModel"

    def __init__(self):
        NoiseModelBase.__init__(self)
        self.nparam = 2
        self.set_init_parameters()

    def set_init_parameters(self, oseries=None):
        self.parameters.loc["noise_alpha"] = (10, 1e-9, np.inf, True, "noise")
        self.parameters.loc["noise_beta"] = (10, 1e-9, np.inf, True, "noise")

    def simulate(self, res, parameters):
        alpha = parameters[0]
        beta = parameters[1]

        # Calculate the time steps
        odelt = (res.index[1:] - res.index[:-1]).values / Timedelta("1d")
        a = self.calculate_noise(res.values, odelt, alpha, beta)
        return Series(index=res.index, data=a, name="Noise")

    @staticmethod
    @njit
    def calculate_noise(res, odelt, alpha, beta):
        # Create an array to store the noise
        a = np.zeros_like(res)
        a[0] = res[0]
        # We have to loop through each value
        for i in range(1, res.size):
            a[i] = res[i] - res[i - 1] * np.exp(-odelt[i - 1] / alpha) - \
                   a[i - 1] * np.exp(-odelt[i - 1] / beta)
        return a



"""The following methods may be used for the diagnostic checking of the
residual time series of a calibrated (Pastas) model.

.. codeauthor:: R.A Collenteur

.. currentmodule:: pastas.stats.tests

.. autosummary::
   :nosignatures:
   :toctree: generated/

    durbin_watson
    ljung_box
    runs_test
    stoffer_toloi
    diagnostics
    plot_acf
    plot_diagnostics

"""

from logging import getLogger

import matplotlib.pyplot as plt
from numpy import sqrt, cumsum, nan, zeros, arange, finfo
from pandas import DataFrame
from scipy.stats import chi2, norm, shapiro, normaltest, probplot

from .core import acf as get_acf

logger = getLogger(__name__)
__all__ = ["durbin_watson", "ljung_box", "runs_test", "stoffer_toloi",
           "diagnostics", "plot_acf", "plot_diagnostics"]


def durbin_watson(series=None):
    """Durbin-Watson test for autocorrelation.

    Parameters
    ----------
    series: pandas.Series, optional
        residuals series

    Returns
    -------
    dw_stat: float
        The method returns the Durbin-Watson test statistic.

    Notes
    -----
    The Durban Watson statistic ([durbin_1951]_, [Fahidy_2004]_) can be used
    to make a statement on the correlation between the values. The formula
    to calculate the Durbin-Watson statistic (DW) is:

    .. math::
        DW = 2 * (1 - \\rho)

    where acf is the autocorrelation of the series for lag s. By
    definition, the value of DW is between 0 and 4. A value of zero
    means complete negative correlation and 4 indicates complete
    positive autocorrelation. A value of zero means no autocorrelation.

    References
    ----------
    .. [durbin_1951] Durbin, J., & Watson, G. S. (1951). Testing for serial
      correlation in least squares regression. II. Biometrika, 38(1/2),
      159-177.

    .. [Fahidy_2004] Fahidy, T. Z. (2004). On the Application of Durbin-Watson
      Statistics to Time-Series-Based Regression Models. CHEMICAL ENGINEERING
      EDUCATION, 38(1), 22-25.

    TODO
    ----
    Compare calculated statistic to critical values, which are problematic
    to calculate and should probably come from a predefined table.

    Examples
    --------
    >>> res = pd.Series(index=pd.date_range(start=0, periods=1000, freq="D"),
    >>>                data=np.random.rand(1000))
    >>>result = ps.stats.durbin_watson(res)

    """
    if not series.index.inferred_freq:
        logger.warning("Caution: The Durbin-Watson test should only be used "
                       "for time series with equidistant time steps.")

    rho = series.autocorr(lag=1)  # Take the first value of the ACF

    dw_stat = 2 * (1 - rho)
    p = nan  # NotImplementedYet
    return dw_stat, p


def ljung_box(series=None, lags=365, nparam=0, full_output=False):
    """Ljung-box test for autocorrelation.

    Parameters
    ----------
    series: pandas.Series, optional
        series to calculate the autocorrelation for that is used in the
        Ljung-Box test.
    lags: int, optional
        The maximum lag to compute the Ljung-Box test statistic for.
    nparam: int, optional
        NUmber of calibrated parameters in the model.
    full_output: bool, optional
        Return the result of the test as a boolean (True) or not (False).

    Returns
    -------
    q_stat: float
        The computed Q test statistic.
    pval: float
        The probability of the computed Q test statistic.

    Notes
    -----
    The Ljung-Box test [Ljung_1978]_ can be used to test autocorrelation in the
    residuals series which are used during optimization of a model. The
    Ljung-Box Q-test statistic is calculated as :

    .. math::
        Q(k) = n * (n + 2) * \\sum(\\frac{\\rho^2(k)}{n - k}

    where $k$ are the lags to calculate the autocorrelation for,
    $n$ is the number of observations and :math:`\\rho(k)` is the
    autocorrelation for lag $k$. The Q-statistic can be compared to the
    value of a Chi-squared distribution to check if the Null hypothesis (no
    autocorrelation) is rejected or not. The hypothesis is rejected when:

    .. math::
        Q(k) > \\chi^2_{\\alpha, h}

    Where :math:`\\alpha` is the significance level and $h$ is the degree of
    freedom defined by $h = n - p$ where $p$ is the number of parameters
    in the model.

    References
    ----------
    .. [Ljung_1978] Ljung, G. and Box, G. (1978). On a Measure of Lack of Fit
      in Time Series Models, Biometrika, 65, 297-303.

    Examples
    --------
    >>> res = pd.Series(index=pd.date_range(start=0, periods=1000, freq="D"),
    >>>                 data=np.random.rand(1000))
    >>> q_stat, pval = ps.stats.ljung_box(res)

    See Also
    --------
    pastas.stats.acf
        This method is called to compute the autocorrelation function.

    """
    if not series.index.inferred_freq:
        logger.warning("Caution: The Ljung-Box test should only be used "
                       "for time series with equidistant time steps. "
                       "Consider using ps.stats.stoffer_toloi instead.")

    acf = get_acf(series, lags=lags, bin_method="regular")
    nobs = series.index.size

    # Drop zero-lag from the acf and drop nan-values as k > 0
    acf = acf.drop(0, errors="ignore").dropna()
    lags = arange(1, len(acf) + 1)

    q_stat = nobs * (nobs + 2) * cumsum(acf.values ** 2 / (nobs - lags))
    dof = max(lags[-1] - nparam, 1)
    pval = chi2.sf(q_stat, df=dof)

    if full_output:
        result = DataFrame(data={"Q Stat": q_stat, "P-value": pval},
                           index=acf.index)
        return result
    else:
        return q_stat[-1], pval[-1]


def runs_test(series, cutoff="mean"):
    """Runs test for autocorrelation.

    Parameters
    ----------
    series: pandas.Series
        Time series to test for autocorrelation.
    cutoff: str or float, optional
        String set to "mean" or "median" or a float to use as the cutoff.

    Returns
    -------
    z_stat: float
        Runs test statistic.
    pval: float
        p-value for the test statistic, based on a normal distribution.

    Notes
    -----
    Distribution free test to check if a time series exhibits significant
    autocorrelation [bradley_1968]_. If :math:`|Z| \\geq Z_{1-\\frac{\\alpha}{
    2}}` then the null hypothesis (Ho) is rejected.

    - Ho: The series is a result of a random process
    - Ha: The series is not the result of a random process

    References
    ----------
    .. [bradley_1968] Bradley, J. V. (1968). Distribution-free statistical
      tests.

    Examples
    --------
    >>> res = pd.Series(index=pd.date_range(start=0, periods=1000, freq="D"),
    >>>                 data=np.random.rand(1000))
    >>> stat, pval = ps.stats.runs_test(res)

    """
    # Make dichotomous sequence
    r = series.values.copy()
    if cutoff == "mean":
        cutoff = r.mean()
    elif cutoff == "median":
        cutoff = r.median()
    elif isinstance(cutoff, float):
        pass
    else:
        raise NotImplementedError("Cutoff criterion {} is not "
                                  "implemented".format(cutoff))

    r[r > cutoff] = 1
    r[r < cutoff] = 0

    # Calculate number of positive and negative noise
    n_pos = r.sum()
    n_neg = r.size - n_pos

    # Calculate the number of runs
    runs = r[1:] - r[0:-1]
    n_runs = sum(abs(runs)) + 1

    # Calculate the expected number of runs and the standard deviation
    n_neg_pos = 2.0 * n_neg * n_pos
    n_runs_exp = n_neg_pos / (n_neg + n_pos) + 1
    n_runs_std = (n_neg_pos * (n_neg_pos - n_neg - n_pos)) / \
                 ((n_neg + n_pos) ** 2 * (n_neg + n_pos - 1))

    # Calculate Z-statistic and pvalue
    z_stat = (n_runs - n_runs_exp) / sqrt(n_runs_std)
    pval = 2 * norm.sf(abs(z_stat))

    return z_stat, pval


def stoffer_toloi(series, lags=365, nparam=0, freq="D"):
    """Adapted Ljung-Box test to deal with missing data [stoffer_1992]_.

    Parameters
    ----------
    series: pandas.Series
        Time series to compute the adapted Ljung-Box statistic for.
    lags: int, optional
        If lags is None, then the default maximum lag is 365. The units of
        this variable depend on the frequency chosen with the freq keyword.
    nparam: int, optional
        Number of parameters of the noisemodel.
    freq: str, optional
        String with the frequency to resample the time series to.

    Returns
    -------
    qm: float
        Adapted Ljung-Box test statistic.
    pval: float
        p-value for the test statistic, based on a chi-squared distribution.

    Notes
    -----
    stoffer-toloi test can handle missing data (nan's) in the input time
    series.

    Reference
    ---------
    .. [stoffer_1992] Stoffer, D. S., & Toloi, C. M. (1992). A note on the
       Ljung—Box—Pierce stoffer_toloi statistic with missing data. Statistics &
       probability letters, 13(5), 391-396.

    Examples
    --------
    >>> res = pd.Series(index=pd.date_range(start=0, periods=1000, freq="D"),
    >>>                data=np.random.rand(1000))
    >>> result = ps.stats.stoffer_toloi(res)

    """
    series = series.asfreq(freq=freq)  # Make time series equidistant

    # TODO: Check if minimum frequency of the is (much) higher than freq and
    #  raise a warning. See also #https://github.com/pastas/pastas/issues/228

    nobs = series.size
    z = (series - series.mean()).fillna(0.0)
    y = z.to_numpy()
    yn = series.notna().to_numpy()

    dz0 = (y ** 2).sum() / nobs
    da0 = (yn ** 2).sum() / nobs
    de0 = dz0 / da0

    # initialize
    dz = zeros(lags)
    da = zeros(lags)
    de = zeros(lags)

    for i in range(0, lags):
        hh = y[:-i - 1] * y[i + 1:]
        dz[i] = hh.sum() / nobs
        hh = yn[:-i - 1] * yn[i + 1:]
        da[i] = hh.sum() / (nobs - i - 1)
        if abs(da[i]) > finfo(float).eps:
            de[i] = dz[i] / da[i]

    re = de / de0
    k = arange(1, lags + 1)
    # Compute the Q-statistic
    qm = nobs ** 2 * sum(da * re ** 2 / (nobs - k))

    dof = max(lags - nparam, 1)
    pval = chi2.sf(qm, df=dof)

    return qm, pval


def diagnostics(series, alpha=0.05, nparam=0, lags=365, stats=(),
                float_fmt="{0:.2f}"):
    """Methods to compute various diagnostics checks for a time series.

    Parameters
    ----------
    series: pandas.Series
        Time series to compute the diagnostics for.
    alpha: float, optional
        significance level to use for the hypothesis testing.
    nparam: int, optional
        Number of parameters of the noisemodel.
    lags: int, optional
        Maximum number of lags (in days) to compute the autocorrelation
        tests for.
    stats: list, optional
        List with the diagnostic checks to perform. Not implemented yet.
    float_fmt: str
        String to use for formatting the floats in the returned DataFrame.

    Returns
    -------
    df: Pandas.DataFrame
        DataFrame with the information for the diagnostics checks.

    Notes
    -----
    Different tests are computed depending on the regularity of the time
    step of the provided time series. series.index.inferred_freq is used to
    determined whether or not the time steps are regular.

    Examples
    --------
    >>> res = pd.Series(index=pd.date_range(start=0, periods=1000, freq="D"),
    >>>                 data=np.random.rand(1000))
    >>> ps.stats.diagnostics(res)
    Out[0]:
                      Checks Statistic P-value  Reject H0
    Shapiroo       Normality      1.00    0.86      False
    D'Agostino     Normality      1.18    0.56      False
    Runs test      Autocorr.     -0.76    0.45      False
    Durbin-Watson  Autocorr.      2.02     nan      False
    Ljung-Box      Autocorr.      5.67    1.00      False

    """
    cols = ["Checks", "Statistic", "P-value"]
    df = DataFrame(index=stats, columns=cols)

    # Shapiroo-Wilk test for Normality
    stat, p = shapiro(series)
    df.loc["Shapiroo", cols] = "Normality", stat, p,

    # D'Agostino test for Normality
    stat, p = normaltest(series)
    df.loc["D'Agostino", cols] = "Normality", stat, p

    # Runs test for autocorrelation
    stat, p = runs_test(series)
    df.loc["Runs test", cols] = "Autocorr.", stat, p

    # Do different tests depending on time step
    if series.index.inferred_freq:
        # Ljung-Box test for autocorrelation
        stat, p = ljung_box(series, nparam=nparam, lags=lags)
        df.loc["Ljung-Box", cols] = "Autocorr.", stat, p

        # Durbin-Watson test for autocorrelation
        stat, p = durbin_watson(series)
        df.loc["Durbin-Watson", cols] = "Autocorr.", stat, p
    else:
        # Stoffer-Toloi for autocorrelation
        stat, p = stoffer_toloi(series, nparam=nparam, lags=lags)
        df.loc["Stoffer-Toloi", cols] = "Autocorr.", stat, p

    df["Reject H0"] = df.loc[:, "P-value"] < alpha
    df[["Statistic", "P-value"]] = \
        df[["Statistic", "P-value"]].applymap(float_fmt.format)

    return df


def plot_acf(series, alpha=0.05, lags=365, acf_options=None, smooth_conf=True,
             ax=None, figsize=(5, 2)):
    """Method to plot the autocorrelation function.

    Parameters
    ----------
    series: pandas.Series
        Residual series to plot the autocorrelation function for.
    alpha: float, optional
        Significance level to calculate the (1-alpha)-confidence intervals.
        For 95% confidence intervals, alpha should be 0.05.
    lags: int, optional
        Maximum number of lags (in days) to compute the autocorrelation for.
    acf_options: dict, optional
        Dictionary with keyword arguments passed on to pastas.stats.acf.
    smooth_conf: bool, optional
        For irregular time series the confidence interval may be
    ax: matplotlib.axes.Axes, optional
        Matplotlib Axes instance to plot the ACF on. A new Figure and Axes
        is created when no value for ax is provided.
    figsize: Tuple, optional
        2-D Tuple to determine the size of the figure created. Ignored if ax
        is also provided.

    Returns
    -------
    ax: matplotlib.axes.Axes

    Examples
    --------
    >>> res = pd.Series(index=pd.date_range(start=0, periods=1000, freq="D"),
    >>>                 data=np.random.rand(1000))
    >>> ps.stats.plot_acf(res)

    """
    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=figsize)

    # Plot the autocorrelation
    if acf_options is None:
        acf_options = {}
    r = get_acf(series, full_output=True, alpha=alpha, lags=lags,
                **acf_options)

    if smooth_conf:
        conf = r.stderr.rolling(10, min_periods=1).mean().values
    else:
        conf = r.stderr.values

    ax.fill_between(r.index.days, conf, -conf, alpha=0.3)
    ax.vlines(r.index.days, [0], r.loc[:, "acf"].values)

    ax.set_xlabel("Lag [Days]")
    ax.set_xlim(0, r.index.days.max())
    ax.set_ylabel('Autocorrelation [-]')
    ax.set_title("Autocorrelation plot")

    ax.grid()
    return ax


def plot_diagnostics(series, alpha=0.05, bins=50, acf_options=None,
                     figsize=(10, 6), **kwargs):
    """Plot a window that helps in diagnosing basic model assumptions.

    Parameters
    ----------
    series: pandas.Series
        Pandas Series with the residual time series to diagnose.
    alpha: float, optional
        Significance level to calculate the (1-alpha)-confidence intervals.
    bins: int optional
        Number of bins used for the histogram. 50 is default.
    acf_options: dict, optional
        Dictionary with keyword arguments passed on to pastas.stats.acf.
    figsize: tuple, optional
        Tuple with the height and width of the figure in inches.
    **kwargs: dict, optional
        Optional keyword arguments, passed on to plt.figure.

    Returns
    -------
    axes: matplotlib.axes.Axes

    Examples
    --------
    >>> res = pd.Series(index=pd.date_range(start=0, periods=1000, freq="D"),
    >>>                 data=np.random.normal(0, 1, 1000))
    >>> ps.stats.plot_diagnostics(res)

    Note
    ----
    The two right-hand side plots assume that the noise or residuals follow a
    Normal distribution.

    See Also
    --------
    pastas.stats.acf
        Method that computes the autocorrelation.
    scipy.stats.probplot
        Method use to plot the probability plot.

    """
    # Create the figure and axes
    fig = plt.figure(figsize=figsize, **kwargs)
    shape = (2, 3)
    ax = plt.subplot2grid(shape, (0, 0), colspan=2, rowspan=1)
    ax1 = plt.subplot2grid(shape, (1, 0), colspan=2, rowspan=1)
    ax2 = plt.subplot2grid(shape, (0, 2), colspan=1, rowspan=1)
    ax3 = plt.subplot2grid(shape, (1, 2), colspan=1, rowspan=1)

    # Plot the residuals or noise series
    ax.axhline(0, c="k")
    series.plot(ax=ax)
    ax.set_ylabel(series.name)
    ax.set_xlim(series.index.min(), series.index.max())
    ax.set_title("{} (n={:.0f}, $\\mu$={:.2f})".format(series.name,
                                                       series.size,
                                                       series.mean()))
    ax.grid()

    # Plot the autocorrelation
    plot_acf(series, alpha=alpha, acf_options=acf_options, ax=ax1)

    # Plot the histogram for normality and add a 'best fit' line
    _, bins, _ = ax2.hist(series.values, bins=bins, density=True)
    y = norm.pdf(bins, series.mean(), series.std())
    ax2.plot(bins, y, 'k--')
    ax2.set_ylabel("Probability density")
    ax2.set_title("Histogram")

    # Plot the probability plot
    probplot(series, plot=ax3, dist="norm", rvalue=True)
    c = ax.get_lines()[1].get_color()
    ax3.get_lines()[0].set_color(c)
    ax3.get_lines()[1].set_color("k")

    plt.tight_layout()
    ax.set_xticklabels(ax.get_xticklabels(), rotation=0, ha="center")

    return fig.axes
